#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

typedef uint8_t byte_t;
typedef uint64_t word_t;

#define WORD_SIZE sizeof(word_t)
#define MIN(A, B) ((A) < (B) ? (A) : (B))
#define MAX(A, B) ((A) > (B) ? (A) : (B))

#ifndef LITTLE_ENDIAN
static const uint32_t __i = 0xFFFF0000;
#define LITTLE_ENDIAN ((*((byte_t *)&__i)) == 0)
#endif

void convert_bytes_le(byte_t *bytes, word_t n)
{
  if (!LITTLE_ENDIAN)
    for (word_t i = 0; i < (n / 2); ++i)
    {
      byte_t temp      = bytes[i];
      bytes[i]         = bytes[n - i - 1];
      bytes[n - i - 1] = temp;
    }
}

void log_result(const char *filename, long size, long msec)
{
  FILE *fp = fopen(filename, "a");
  fprintf(fp, "%lu, %lu\n", size, msec);
  fclose(fp);
}

void fill_buffer(byte_t *bytes, word_t size)
{
  for (word_t i = 0; i < size; ++i)
    bytes[i] = rand() % 255;
}

void iterate_1(byte_t *a, byte_t *b, word_t size, byte_t *ret)
{
  for (word_t i = 0; i < size; ++i)
  {
    ret[i] = a[i] + b[i] + ret[size];
    if ((UINT8_MAX - a[i]) < b[i] || (UINT8_MAX - a[i] - ret[size]) < b[i])
      ret[size] = 1;
    else
      ret[size] = 0;
  }
}

void iterate_2(byte_t *a, byte_t *b, word_t size, byte_t *ret)
{
  const word_t bound =
      size % WORD_SIZE == 0 ? (size / WORD_SIZE) : (size / WORD_SIZE) + 1;
  for (word_t i = 0; i < bound; ++i)
  {
    const word_t index = WORD_SIZE * i;
    const word_t _size = MIN(WORD_SIZE, size - index);
    word_t a_i = 0, b_i = 0;
    memcpy(&a_i, a + index, _size);
    memcpy(&b_i, b + index, _size);
    word_t res = a_i + b_i + ret[size];
    memcpy(ret + index, &res, _size);
    if ((((1LU << (8 * _size)) - 1) - a_i) < b_i ||
        (((1LU << (8 * _size)) - 1) - a_i - ret[size]) < b_i)
      ret[size] = 1;
    else
      ret[size] = 0;
  }
}

#ifndef TYPE
#define TYPE 1
#endif

int main()
{
  srand(time(NULL));
  const word_t trials   = 10;
  const word_t orders[] = {
      20, 21, 22, 23, 24, 25, 26, 27, 28, 29,
  };
  for (word_t i = 0; i < sizeof(orders) / sizeof(orders[0]); ++i)
  {
    const word_t size = 1LU << orders[i];
    byte_t *a         = calloc(1, size);
    byte_t *b         = calloc(1, size);
    byte_t *ret       = calloc(1, size + 1);
    for (word_t j = 0; j < trials; ++j)
    {
      printf("{%d}[2^%lu/%lu]: Generating buffers\n", TYPE, orders[i], j);
      memset(ret, 0, size + 1);
      fill_buffer(a, size);
      fill_buffer(b, size);
      printf("{%d}[2^%lu/%lu]: Generated buffers\n", TYPE, orders[i], j);

      clock_t start = clock(), diff;
#if TYPE == 1
      iterate_1(a, b, size, ret);
#else
      iterate_2(a, b, size, ret);
#endif
      diff      = clock() - start;
      long msec = (diff * 1000) / CLOCKS_PER_SEC;
      printf("{%d}[2^%lu/%lu]: Finished computation\n\n", TYPE, orders[i], j);
#if TYPE == 1
      log_result("iterate_1", orders[i], msec);
#else
      log_result("iterate_2", orders[i], msec);
#endif
    }
    free(a);
    free(b);
    free(ret);
    printf("\n");
  }
  return 0;
}
